@@ -1750,9 +1750,16 @@ TOOLS_VERSION_UINT_TO_COMPONENTS(const ToolsVersion toolsVersion,   // IN
 #define   TOOLS_VERSION_BANDSAW_UPDATE1_V_MNR  2
 #define   TOOLS_VERSION_BANDSAW_UPDATE1_V_BASE 5
 
-#define   TOOLS_VERSION_CURRENT        TOOLS_VERSION_BANDSAW_UPDATE1
-#define   TOOLS_VERSION_CURRENT_STR    TOOLS_VERSION_TO_STR(TOOLS_VERSION_BANDSAW_UPDATE1)
-#define   TOOLS_VERSION_CURRENT_CSV    TOOLS_VERSION_TO_CSV(TOOLS_VERSION_BANDSAW_UPDATE1)
+#ifndef RC_INVOKED
+#define   TOOLS_VERSION_NEXT TOOLS_VERSION_TO_UINT(TOOLS_VERSION_NEXT_V)
+#endif /* RC_INVOKED */
+#define   TOOLS_VERSION_NEXT_V_MJR  12
+#define   TOOLS_VERSION_NEXT_V_MNR  3
+#define   TOOLS_VERSION_NEXT_V_BASE 0
+
+#define   TOOLS_VERSION_CURRENT        TOOLS_VERSION_NEXT
+#define   TOOLS_VERSION_CURRENT_STR    TOOLS_VERSION_TO_STR(TOOLS_VERSION_NEXT)
+#define   TOOLS_VERSION_CURRENT_CSV    TOOLS_VERSION_TO_CSV(TOOLS_VERSION_NEXT)
 
 /*
  * The extended Tools version is the current Tools version with the
